/*
 * Sat Jan 31 10:55:55 CET 2004
 * vecna@s0ftpj.org
 *
 * example of steganography simulating common (worm) traffic,
 *
 * http://www.securityfocus.com/archive/75/335132/2003-08-21/2003-08-27/0
 * http://cert.uni-stuttgart.de/archive/incidents/2003/10/msg00143.html
 *
 * this code is part of a BFi article, go to www.s0ftpj.org and
 * get a lot of information about this.
 *
 * this is blaststegd and it has to work with blaststegsender,
 *
 * this blaststegd could listen for a lot of sessions and make a file for each
 * one with the session dump
 *
 * the client sends anonymous hidden data over tcp reset packet commonly
 * generated for blastmer workaround
 *
 * the session inside the caos is discriminated with a pre-shared key
 * the format of file required from blaststegd is a simple list of
 * pre-shared key.
 *
 * compiled under linux, gcc file.c -o output
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include <linux/if_ether.h>
#include <linux/if_packet.h>

#include "blaststeg.h"

struct session
{
  char *key;
  unsigned char stream[STREAMSIZE];
  unsigned short port;
  unsigned int counter, length, readbyte;
  FILE *dump;
};
struct session *tracked;

int
main (int argc, char **argv)
{
  FILE *key;
  char line[MAXKEYLEN], packet[MAXRSTSIZE], *dump_directory;
  int sockfd, i = 0, list_sess = 0;

  if (argc < 2)
    {
      printf ("%s pre-shared-password-file <optional-dump-dir>\n", *argv);
      exit (1);
    }

  if (argc == 3)
    dump_directory = argv[2];
  else
    dump_directory = strdup ((const char *)get_current_dir_name ());

  printf (			/* BANNER! :) SPAM && INFO */
   "KEEP YOUR PRIVACY! this is a free software for communication hiding\n"
   "blaststeg daemon, anonymous steganographed packets receiver\n"
   "coded on Sat Jan 31 2004 vecna@s0ftpj.org, http://www.s0ftpj.org\n"
   "READ ALL ABOUT THIS SOFTWARE, I'm acting like worm-traffic emulation...\n"
   "if a lot of time is past since the time i coded this it could be obsolete\n"
   "and insecure!\n"
   "this software could make some complete/incomplete dump on %s\n"
   "don't forget about that\n\n", dump_directory);

  if ((key = fopen (argv[1], "r")) == NULL)
    {
      printf ("unable to open file %s\n", argv[1]);
      exit (1);
    }

  do
    {
      fgets (line, MAXKEYLEN, key);
      list_sess++;

    }
  while (!feof (key));

  if ((tracked = (void *) calloc (list_sess, sizeof (*tracked))) == NULL)
    {
      printf ("unable to alloc memory\n");
      exit (1);
    }

  rewind (key);

  do
    {
      fgets (line, MAXKEYLEN, key);

      /* strip '\n' */
      line[strlen (line) - 1] = 0x00;

      tracked[i].key = strdup (line);
      compute_stream (tracked[i].stream, line);
      i++;

    }
  while (!feof (key));

  if ((sockfd = socket (PF_PACKET, SOCK_DGRAM, htons(ETH_P_IP))) == -1)
    {
      printf ("unable to open datalink layer socket\n");
      exit (1);
    }

  while (read (sockfd, packet, MAXRSTSIZE) != -1)
    {
      static char fname[MAXPATHLEN], newname[MAXPATHLEN];
      struct iphdr *ip = (struct iphdr *) packet;
      struct tcphdr *tcp = (struct tcphdr *) (packet + sizeof (*ip));

      if (!tcp->rst || !tcp->ack)
	continue;

      if (ip->ihl != 5 || tcp->doff != 5)
	continue;

/* search if is a new session */
      for (i = 0; i < list_sess; i++)
	{
	  if (!memcmp (tracked[i].stream, &tcp->ack_seq, 4))
	    {

	      printf ("new [%s] session detected\n", tracked[i].key);

	      sprintf (fname, "%s/%s-%u-%s",
	         dump_directory, tracked[i].key, tcp->dest, ACTVSTR
	      );

	      if (tracked[i].dump != NULL)
		{
		  printf ("tuncation of incomplete session %s-%u-%s\n",
			  tracked[i].key, tracked[i].port, ACTVSTR);

		  fclose (tracked[i].dump);
		}

	      if ((tracked[i].dump = fopen (fname, "w+")) == NULL)
		{
		  printf ("unable to open dump %s\n", fname);
		  exit (0);
		}

	      tracked[i].port = tcp->dest;
	      tracked[i].length = ip->id;
	      tracked[i].counter =4;

	      break;

	    } /* new session check */

	  /* match session continuation */
	  if (tracked[i].port == tcp->dest)
	    {
	      unsigned char check;

	      memcpy (&check, &ip->id, 1);

	      if (check == tracked[i].stream[tracked[i].counter])
		{
		  fwrite ((char *) (&ip->id) + 1, 1, 1, tracked[i].dump);
		  fwrite ((char *) &tcp->ack_seq, 1, 4, tracked[i].dump);

		  if (++(tracked[i].counter) == STREAMSIZE)
		    tracked[i].counter = 0;

		  tracked[i].readbyte += 5;

		  if (tracked[i].readbyte < tracked[i].length)
		    break;
	       /* else: session is finished */

		  sprintf (fname, "%s/%s-%u-%s",
			   dump_directory, tracked[i].key, tracked[i].port,
			   ACTVSTR);
		  sprintf (newname, "%s/%s-%u", dump_directory, tracked[i].key,
			   tracked[i].port);

		  if (tracked[i].readbyte > tracked[i].length)
		    fseek (tracked[i].dump, tracked[i].length, SEEK_SET);
	       /* else: readbyte == length and doesn't required fseek */

		  fclose (tracked[i].dump);
		  rename (fname, newname);

		  printf ("session closed and saved on %s (%d byte)\n",
			  newname, tracked[i].length);

		  tracked[i].port = 0;
		  tracked[i].dump = NULL;
		  tracked[i].counter = tracked[i].length =
		    tracked[i].readbyte = 0;
		}
	    }
	}	/* for rolling over tracked[] */
    }	/* while read */

/* never reached if read don't make error */
  printf ("error reading at raw sock layer\n");
  exit (1);
}
