/*
 * Sat Jan 31 10:55:55 CET 2004
 * vecna@s0ftpj.org
 *
 * example of steganography simulating common (worm) traffic,
 *
 * http://www.securityfocus.com/archive/75/335132/2003-08-21/2003-08-27/0
 * http://cert.uni-stuttgart.de/archive/incidents/2003/10/msg00143.html
 *
 * this code is part of a BFi article, go to www.s0ftpj.org and 
 * get a lot of information about it.
 *
 * this is blaststegclient and it has to work with blaststegd
 *
 * this client sends anonymous hidden data over tcp reset packet commonly
 * generated for blastmer workaround
 *
 * blaststegd could listen for a lot of sessions and make a file for each one
 * with the session dump 
 *
 * the session inside the caos is discriminated with a pre-shared key
 *
 * protocol to send packets must respect some things:
 * - the first packet of a session must have tcp->ack matching and ip->id
 *   contains the length of complete data 
 * - the destination port is kept as tracking system, when is matched a 
 * packet with that port it checks if the stream is followed and if it is 
 * then the ack field and one byte of id is kept as incoming data.
 * each packet contains 5 bytes of data and 3 bytes for session tracking.
 */
#define _GNU_SOURCE
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <sys/param.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include "blaststeg.h"

unsigned int
getrand (unsigned int base, unsigned int min, unsigned int max)
{
  unsigned int ret, diff;

  diff = (max - min);

  srandom (time (NULL) + diff);

  if ((ret = random ()) != 0)
    ret %= diff;

  return (base + ret);
}

inline unsigned int
half_cksum (const unsigned short *data, int len)
{
  unsigned int sum = 0x00;
  unsigned short carry = 0x00;

  while (len > 1)
    {
      sum += *data++;
      len -= 2;
    }

  if (len == 1)
    {
      *((unsigned short *) &carry) = *(unsigned char *) data;
      sum += carry;
    }

  return sum;
}

inline unsigned short
compute_sum (unsigned int sum)
{
  sum = (sum >> 16) + (sum & 0xffff);
  sum += (sum >> 16);

  return (unsigned short) ~sum;
}

void
send_packet (int fd, struct iphdr *ip, struct tcphdr *tcp)
{
  unsigned int sum;
  struct sockaddr_in sa;

  sa.sin_addr.s_addr = ip->daddr;
  sa.sin_port = tcp->dest;
  sa.sin_family = PF_INET;

  /* ip check */
  ip->check = 0;
  sum = half_cksum ((unsigned short *) ip, sizeof (struct iphdr));
  ip->check = compute_sum (sum);

  /* tcp check */
  tcp->check = 0;
  sum = half_cksum ((unsigned short *) &ip->saddr, 8);
  sum += htons (IPPROTO_TCP + sizeof(struct tcphdr)); 
  sum += half_cksum ((unsigned short *) tcp, sizeof (struct tcphdr));
  tcp->check = compute_sum (sum);

  if ((sendto
       (fd, (void *) ip, MAXRSTSIZE, 0, (struct sockaddr *) &sa,
	sizeof (sa))) == -1)
    {
      printf ("unable to send sock raw packet!\n");
      exit (1);
    }
}

int
main (int argc, char **argv)
{
  unsigned char stream[STREAMSIZE], packet[MAXRSTSIZE];
  unsigned int counter = 0, delay, fd, hdrincl = 1;
  int filelen;
  struct iphdr *ip = (struct iphdr *) packet;
  struct tcphdr *tcp = (struct tcphdr *) (packet + sizeof (struct iphdr));
  FILE *source;

  if (argc != 4)
    {
      printf ("%s data_file session_key dest_host\n", *argv);
      exit (1);
    }

  printf ("PRIVACY PROTECTION SOFTWARE - example of hiding data on\n"
	  "apparently common worm traffic error - www.s0ftpj.org\n"
	  "check about other information, it could be useful to understand\n"
	  "the limits, the working system and the motivation before run this\n"
	  "steganographic software. coded by vecna@s0ftpj.org\n");

  if ((source = fopen (argv[1], "r")) == NULL)
    {
      printf ("unable to open file %s\n", argv[1]);
      exit (1);
    }

  fseek (source, 0, SEEK_END);
  filelen = ftell (source);
  rewind (source);

  compute_stream (stream, argv[2]);

  if ((fd = socket (PF_INET, SOCK_RAW, IPPROTO_TCP)) == -1)
    {
      printf ("unable to open socket raw (are you root ?)\n");
      exit (1);
    }

  setsockopt (fd, SOL_IP, IP_HDRINCL, &hdrincl, sizeof (int));

  /* default ip hdr */
  memset ((void *) ip, 0x00, sizeof (struct iphdr));
  ip->saddr = inet_addr ("127.0.0.1");
  ip->daddr = inet_addr (argv[3]);
  ip->ihl = 5;
  ip->version = 4;
  ip->protocol = IPPROTO_TCP;
  ip->tot_len = htons (MAXRSTSIZE);
  ip->ttl = (unsigned char) getrand (124, 0, 6);

  /* default tcp hdr */
  memset ((void *) tcp, 0x00, sizeof (struct tcphdr));
  tcp->doff = 5;
  tcp->rst = 1;
  tcp->ack = 1;
  tcp->source = htons (80);
  tcp->dest = htons ((unsigned short) getrand (1024, 0, 920));

  /* initialization ip settings */
  ip->id = filelen;

  /* initialization tcp settings */
  memcpy ((unsigned char *) &tcp->ack_seq, &stream[counter], 4);
  counter += 4;

  send_packet (fd, ip, tcp);

  while (1)
    {
      delay = getrand (BASEDELAY, 0, MAXDELAY);
      sleep (delay);

      ip->ttl = (unsigned char) getrand (124, 0, 6);

      memcpy (&ip->id, &stream[counter], 1);

      fread ((unsigned char *) &ip->id + 1, 1, 1, source);

      if (filelen < 4)
	tcp->ack_seq = getrand (0, 123, 4000000);

      fread ((void *) &tcp->ack_seq, 4, 1, source);

      send_packet (fd, ip, tcp);

      if (++counter == STREAMSIZE)
	counter = 0;

      filelen -= 5;
      if (filelen < 0)
	break;
    }

  printf ("file sent\n");
  return 0;
}
